/*
 * Copyright 1998-2001, University of Notre Dame.
 * Authors: Jeffrey M. Squyres, Arun Rodrigues, and Brian Barrett with
 *          Kinis L. Meyer, M. D. McNally, and Andrew Lumsdaine
 * 
 * This file is part of the Notre Dame LAM implementation of MPI.
 * 
 * You should have received a copy of the License Agreement for the Notre
 * Dame LAM implementation of MPI along with the software; see the file
 * LICENSE.  If not, contact Office of Research, University of Notre
 * Dame, Notre Dame, IN 46556.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted subject to the conditions specified in the
 * LICENSE file.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * Additional copyrights may follow.
 * 
 *	Ohio Trollius
 *	Copyright 1997 The Ohio State University
 *	GDB/NJN
 *
 *	$Id: kcreate.c,v 1.6 2001/01/22 22:51:26 jsquyres Exp $
 *
 *	Function:	- creates an OTB process
 *			- does not make a Trollius kernel request
 *			- essentially a packaged fork and exec
 *			- redirects created process's stdio
 *	Accepts:	- child program name
 *			- argument vector
 *			- environment vector
 *			- stdio file descriptors (0 if no redirection)
 *	Returns:	- (to parent) child pid or ERROR
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "kio.h"
#include "preq.h"
#include "terror.h"
#include "typical.h"
#include "lamlog.h"

/*
 * external functions
 */
extern void		_cipc_cleanup();
extern int		lam_pty_open();
extern int		liosattach();
extern int		liofdattach();
extern int4		stoi();

/*
 * private functions
 */
static int		redirect_stdio(int, int *);
static int		set_environment(char **);


int
kcreate(const char *pathname, char **argv)
{
    int			pid;

    if ((pid = fork()) < 0)
	return(LAMERROR);
    else if (pid == 0) {
	_cipc_cleanup();
	execvp(pathname, argv);
	exit(errno);
    }

    return(pid);
}

int
kcreatev(const char *pathname, char **argv, char **envv, char *wrkdir, 
	 int *fds)
{
    struct sigaction	act;
    sigset_t		sigs;
    int			pid;
    int			rtf;
    char		*p;

    if ((pid = fork()) < 0)
	return(LAMERROR);
    else if (pid == 0) {
/*
 * Unblock all signals.
 */
	sigprocmask(0, 0, &sigs);
	sigprocmask(SIG_UNBLOCK, &sigs, 0);
/*
 * Reset SIGCHLD and SIGPIPE handlers.
 */
	act.sa_handler = SIG_DFL;
	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaction(SIGCHLD, &act, 0);
	sigaction(SIGPIPE, &act, 0);

	(void) setsid();
/*
 * Redirect the stdio fd's
 */
	rtf = (p = getenv("TROLLIUSRTF")) ? stoi((unsigned char *) p) : 0;
	if (redirect_stdio(rtf, fds))
	    exit(errno);
/*
 * close all the rest of the file descriptors.  The for loop may seem
 * redundant, but don't forget the case where the lamd forks a new
 * process.  All the sockets to child connections that the lamd has
 * open will still be open.  The file descriptors for these sockets
 * are not stored anywhere other than an internal fd_set in
 * otb/sys/kernel/kernelio.c.  PLUS, since this routine is
 * [potentially] called from within LAM/MPI (i.e., outside of the
 * lamd), we don't want to make any lamd-specific things happen here.
 *
 * The for loop should be ok -- we really don't want to pass on any
 * file descriptors other than stdin/stdout/stderr, and those get
 * passed below.  
 */
	_cipc_cleanup();
	lamcloselog();
#if 0
	/* This should no longer be necessary -- the file descriptor
           leak has been fixed properly.  otb/sys/kernel/kernelio.c
           has now been set to make new file descriptor connections to
           be close-on-exec so that they won't be inherited by this
           new process. */
	for (i = 3; i < FD_SETSIZE; i++)
	  close(i);
#endif

	if (set_environment(envv))
	    exit(errno);

	if (wrkdir) {
	    if (chdir(wrkdir) && !(rtf & RTF_TRYCWD)) {
		fprintf(stderr, "kcreatev: chdir(%s): ", wrkdir);
		terror("");
		exit(errno);
	    }
	}

	execvp(pathname, argv);
	exit(errno);
    }

    return(pid);
}

/*
 *	redirect_stdio
 *
 *	Function:	- redirect process stdio
 *	Accepts:	- runtime flags
 *			- passed file descriptors or NULL
 */
static int
redirect_stdio(int rtf, int *fds)
{
    int			ionode;
    int			stdout_fd;
    int			stderr_fd;
    int			master;
    int			slave;
    char		*p;

    if (fds) {
/*
 * Redirect stdio to the passed fds.
 */
	dup2(fds[0], fileno(stdin));
	dup2(fds[1], fileno(stdout));
	dup2(fds[2], fileno(stderr));
    } else {
	if ((rtf&RTF_IO) && !(rtf&RTF_PFDIO) && (p = getenv("TROLLIUSFD"))) {
/*
 * Connect UNIX stderr and stdout to IO daemon.
 */
	    ionode = stoi((unsigned char *) p);
	    p = strchr(p, ':') + 1;
	    p = strchr(p, ':') + 1;			/* skip stdin */
	    stdout_fd = stoi((unsigned char *) p);
	    p = strchr(p, ':') + 1;
	    stderr_fd = stoi((unsigned char *) p);
/*
 * Attach stderr stream to iod nobuffered.
 */
	    if (liosattach(stderr, ionode, stderr_fd, _IONBF))
		return(LAMERROR);
/*
 * Try to use a pty for stdout if asked for and if stdout is a tty.
 */
	    if ((rtf & RTF_PTYS) && (rtf & RTF_TTYOUT)) {
		if (lam_pty_open(&master, &slave)) {
		    if (errno == ENOPTYSUPPORT) {
/*
 * No pty support on this node so fall back to no pty IO.
 */
			if (liosattach(stdout, ionode, stdout_fd, _IONBF))
			    return(LAMERROR);
		    }
		} else {
		    if (dup2(slave, fileno(stdout)) != fileno(stdout))
			return(LAMERROR);
		    if (liofdattach(master, ionode, stdout_fd))
			return(LAMERROR);
		}
	    } else {
		if (liosattach(stdout, ionode, stdout_fd, _IONBF))
		    return(LAMERROR);
	    }
	}
    }

    return(0);
}

/*
 *	set_environment
 *
 *	Function:	- set process environment
 *	Accepts:	- environment vector
 */
static int
set_environment(char **env)
{
    int	i;

    if (env) {
	for (i = 0; env[i]; i++) {
	    if (putenv(env[i]))
		return(LAMERROR);
	}
    }

    return(0);
}
