/*
 * Copyright 1998-2001, University of Notre Dame.
 * Authors: Jeffrey M. Squyres, Arun Rodrigues, and Brian Barrett with
 *          Kinis L. Meyer, M. D. McNally, and Andrew Lumsdaine
 * 
 * This file is part of the Notre Dame LAM implementation of MPI.
 * 
 * You should have received a copy of the License Agreement for the Notre
 * Dame LAM implementation of MPI along with the software; see the file
 * LICENSE.  If not, contact Office of Research, University of Notre
 * Dame, Notre Dame, IN 46556.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted subject to the conditions specified in the
 * LICENSE file.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * Additional copyrights may follow.
 * 
 *	Ohio Trollius
 *	Copyright 1997 The Ohio State University
 *	NJN
 *
 *	$Id: tsetname.c,v 1.2 2001/02/03 20:10:16 jsquyres Exp $
 *
 *	Function:	- set datatype name
 *	Accepts:	- datatype
 *			- datatype name
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <string.h>

#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>


/*@
   MPI_Type_set_name - Set a name on a datatype

Input Parameters:
+ type - MPI datatype
- name - Pointer to the name of dataname

Notes:

This function associates a name with a datatype, which can be
useful for debugging purposes.  The function 'MPI_Type_get_name' can
be used to retrieve name that was previously set with
'MPI_Type_set_name'.

Named predefined datatypes have the default names of the datatype
name. For example, 'MPI_WCHAR' has the default name of `MPI_WCHAR`.
It is legal to change any of these names with calls to
'MPI_Type_set_name'.

Names passed in C must include the null terminator, and can be of
maximum length 'MPI_MAX_OBJECT_NAME' (including the null terminator).
Names passed in Fortran can also have a maximum length of
'MPI_MAX_OBJECT_NAME'.

Setting a name on a datatype that already has a name associated
with it replaces the old name.

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE
.N MPI_ERR_ARG

.seealso MPI_Comm_set_name, MPI_Comm_get_name, MPI_Type_get_name, MPI_Win_get_name, MPI_Win_set_name

.N WEB
@*/
int
MPI_Type_set_name(MPI_Datatype type, char *name)
{
	lam_initerr_m();
	lam_setfunc_m(BLKMPITSETNAME);
/*
 * Check the arguments.
 */
	if (type == MPI_DATATYPE_NULL) {
		return(lam_errfunc(MPI_COMM_WORLD,
			BLKMPITSETNAME, lam_mkerr(MPI_ERR_TYPE, 0)));
	}

	if (name == 0) {
		return(lam_errfunc(MPI_COMM_WORLD,
			BLKMPITSETNAME, lam_mkerr(MPI_ERR_ARG, 0)));
	}

	strncpy(type->dt_name, name, MPI_MAX_OBJECT_NAME);
	type->dt_name[MPI_MAX_OBJECT_NAME - 1] = 0;

	lam_resetfunc_m(BLKMPITSETNAME);
	return(MPI_SUCCESS);
}
