/*
 * Copyright 1998-2001, University of Notre Dame.
 * Authors: Jeffrey M. Squyres, Arun Rodrigues, and Brian Barrett with
 *          Kinis L. Meyer, M. D. McNally, and Andrew Lumsdaine
 * 
 * This file is part of the Notre Dame LAM implementation of MPI.
 * 
 * You should have received a copy of the License Agreement for the Notre
 * Dame LAM implementation of MPI along with the software; see the file
 * LICENSE.  If not, contact Office of Research, University of Notre
 * Dame, Notre Dame, IN 46556.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted subject to the conditions specified in the
 * LICENSE file.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * Additional copyrights may follow.
 * 
 *	Ohio Trollius
 *	Copyright 1997 The Ohio State University
 *	RBD
 *
 *	$Id: unpack.c,v 6.4 1999/12/31 21:27:42 jsquyres Exp $
 *
 *	Function:	- unpack to data buffer
 *	Accepts:	- packed buffer
 *			- packed buffer size
 *			- packed buffer position (in-out)
 *			- data buffer
 *			- data count
 *			- data type
 *			- communicator
 *	Returns:	- MPI_SUCCESS or error code
 */

#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>

/*@
    MPI_Unpack - Unpack a datatype into contiguous memory

Input Parameters:
+ packbuf - input buffer start (choice) 
. packsize - size of input buffer, in bytes (integer) 
. count - number of items to be unpacked (integer) 
. dtype - datatype of each output data item (handle) 
- comm - communicator for packed message (handle) 

Input/Output Parameters:
. packpos - current position in bytes (integer) 

Output Parameter:
. buf - output buffer start (choice) 

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_ARG
.N MPI_ERR_BUFFER
.N MPI_ERR_TRUNCATE

.seealso: MPI_Pack, MPI_Pack_size

.N ACK
@*/
int MPI_Unpack(void *packbuf, int packsize, int *packpos, 
	       void *buf, int count, MPI_Datatype dtype, 
	       MPI_Comm comm)
{
	int		size;
	char		*p;

	lam_initerr();
	lam_setfunc(BLKMPIUNPACK);
/*
 * Check the arguments.
 */
	if (comm == MPI_COMM_NULL) {
		return(lam_errfunc(MPI_COMM_WORLD,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_COMM, 0)));
	}

	if (dtype == MPI_DATATYPE_NULL || (!dtype->dt_commit)) {
		return(lam_errfunc(comm,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_TYPE, 0)));
	}

	if (count < 0) {
		return(lam_errfunc(comm,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_COUNT, 0)));
	}

	if ((packsize < 0) || (packpos == 0)) {
		return(lam_errfunc(comm,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_ARG, 0)));
	}

	if ((*packpos < 0) || ((*packpos >= packsize) && (count > 0))) {
		return(lam_errfunc(comm,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_ARG, 0)));
	}

	if ((packsize > 0) && (packbuf == 0)) {
		return(lam_errfunc(comm,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_BUFFER, 0)));
	}
/*
 * Unpack the data.
 */
	p = ((char *) packbuf) + *packpos;
	size = packsize - *packpos;

	size = lam_unpack(p, size, (char *) buf, count, dtype);

	if (size < 0) {
		return(lam_errfunc(comm,
			BLKMPIUNPACK, lam_mkerr(MPI_ERR_TRUNCATE, 0)));
	}

	*packpos += size;

	lam_resetfunc(BLKMPIUNPACK);
	return(MPI_SUCCESS);
}
